#include "antiglut.h"
#include <math.h>
#include <stdio.h>
#include <assert.h>

#include <mvlib.h>

#include <../source/mvlib.c>
#include <../source/rg.c>
#include <../source/agmainloop.c>

#define MAXFIB	100000
#define MAXC	(10 * MAXFIB)

#define	DRAW_SCALE	(vec_t)5e6

/*
 * Size of simulation box
 */
#define BOXEXT	(vec_t)1e-6
vec3_t bbox[2] = {{-BOXEXT,-BOXEXT,-BOXEXT},{BOXEXT,BOXEXT,BOXEXT}};

int		num_fibers = 10000;
vec_t	fiber_len = 2e-7;
vec_t	fiber_diameter = 2e-9;

vec3_t	fib[MAXFIB][2];

typedef struct fc_info_s
{
	int fn1, fn2;	/* indices of contacting fibers */
	vec_t dist;		/* minimal distance between fibers */
	vec3_t cp;		/* contact point */
}	fc_info_t, *fc_info_p;

int	num_contacts = 0;
fc_info_t fc[MAXC];

static ran_state_t state;

static void
fill_box_with_fibers()
{
	int i;
	vec_t hlen = 0.5f * fiber_len;
	vec3_t fo, halfdir;
	for(i = 0;i < num_fibers;i++)
	{
		ranNextVec3Range(&state,fo,-BOXEXT,BOXEXT);
		ranDirection(&state,halfdir);
		vec3Scale(halfdir,halfdir,hlen);
		vec3Sub(fib[i][0],fo,halfdir);
		vec3Add(fib[i][1],fo,halfdir);
	}
}

/* volumetric rate of fibers */
static vec_t fiber_vrate()
{
	vec_t fvol, vol;
	vec3_t bs;
	vec3Sub(bs, bbox[1], bbox[0]);
	vol = bs[0] * bs[1] * bs[2];
	fvol = num_fibers * M_PI * fiber_diameter * fiber_diameter * (vec_t)0.25 * fiber_len;
	return fvol / vol;
}

static int
check_for_contacts()
{
	int i, j;
	vec3_t cp;
	vec_t cd;
	num_contacts = 0;
	for(i = 0;i < num_fibers;i++)
		for(j = i + 1;j < num_fibers;j++)
		{
			if(!segCrossing(fib[i][0],fib[i][1],fib[j][0],fib[j][1],&cd,cp))
				continue;
			if(cd > fiber_diameter)
				continue;
			if(num_contacts >= MAXC)
			{
				puts("Too many contacts!");
				return num_contacts;
			}
			fc[num_contacts].fn1 = i;
			fc[num_contacts].fn2 = j;
			fc[num_contacts].dist = cd;
			vec3Copy(fc[num_contacts].cp, cp);
			num_contacts++;
		}
	return num_contacts;
}

/*
int		nums_fibers[] = {500,1000,5000};
vec_t	fiber_lens[] = {100e-9,200e-9};
vec_t	fiber_diameters[] = {2e-9,20e-9};

*/

int		nums_fibers[] = {100,500,1000,2000,5000,10000,20000,50000,100000};
vec_t	fiber_lens[] = {50e-9,100e-9,200e-9};
vec_t	fiber_diameters[] = {1e-9,2e-9,5e-9,10e-9,20e-9,40e-9};

void onerun()
{
	char *ptr, buf[1024];
	fill_box_with_fibers();
	check_for_contacts();
	sprintf(buf,"%i;%g;%g;%g;%i",
			num_fibers,
			fiber_len,fiber_diameter,
			fiber_vrate(),
			num_contacts);
	for(ptr = buf;*ptr;ptr++)
		if(*ptr == '.')
			*ptr = ',';
	puts(buf);
	fflush(stdout);
}

void longrun()
{
	int i, j, k;
	int im = sizeof(nums_fibers) / sizeof(int);
	int jm = sizeof(fiber_lens) / sizeof(vec_t);
	int km = sizeof(fiber_diameters) / sizeof(vec_t);
	printf("Starting longrun: %i total iterations\n", im * jm * km);
	for(i = 0;i < im;i++)
	{
		num_fibers = nums_fibers[i];
		assert(num_fibers <= MAXFIB);
		for(j = 0;j < jm;j++)
		{
			fiber_len = fiber_lens[j];
			puts("num_fibers;fiber_len;fiber_diameter;fiber_vrate;num_contacts");
			for(k = 0;k < km;k++)
			{
				fiber_diameter = fiber_diameters[k];
				onerun();
			}
			/* puts("INNER COMPLETE"); */
		}
		/* puts("MID COMPLETE"); */
	}
	/* puts("OUTER COMPLETE"); */
}

vboolean draw_main(void)
{
	int i;
	glPushMatrix();
	glScalef(DRAW_SCALE, DRAW_SCALE, DRAW_SCALE);

	glColor4f(.3f,0.2f,.25f,.5f);
	glLineWidth(1.8f);
	bboxDraw(bbox[0], bbox[1]);

	glLineWidth(1.2f);
	glColor4f(.6f,1,1,.3f);
	glBegin(GL_LINES);
	for(i = 0;i < num_fibers;i++)
	{
		glVertex3fv(fib[i][0]);
		glVertex3fv(fib[i][1]);
	}
	glEnd();

	glPointSize(2.5f);
	glColor4f(0,0,0,.8f);
	glBegin(GL_POINTS);
	for(i = 0;i < num_contacts;i++)
		glVertex3fv(fc[i].cp);
	glEnd();
	glLineWidth(1.4f);
	glColor4f(0.7f,0,0.4f,.2f);
	glBegin(GL_LINES);
	for(i = 0;i < num_contacts;i++)
	{
		glVertex3fv(fib[fc[i].fn1][0]);
		glVertex3fv(fib[fc[i].fn1][1]);
		glVertex3fv(fib[fc[i].fn2][0]);
		glVertex3fv(fib[fc[i].fn2][1]);
	}
	glEnd();
	
	glPopMatrix();
	return true;
}


int main()
{
	assert(sizeof(GLfloat) == sizeof(vec_t));
	ranInitState(&state,1);
	/*
	longrun();
	*/
	fill_box_with_fibers();
	check_for_contacts();
	printf("num_fibers %i;length %g;diameter %g;vf %g;number of contacts %i\n",
			num_fibers,
			fiber_len,fiber_diameter,
			fiber_vrate(),
			num_contacts);

	agWindowConfig("TEST",800,800,GL_FALSE);
	if(!agInit(AG_REQUIRED_1_1,GL_FALSE,GL_FALSE,GL_FALSE))
		return -1;

	glClearColor(1,0.95f,0.8f,0);

	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	
	agMainLoop(draw_main, NULL);
	agShutdown();

	return 0;
}
